@@ -16,27 +16,30 @@
 
 .. versionadded:: 2.7
 """
+import copy
+
 from itertools import islice
 
 from bson.objectid import ObjectId
 from bson.raw_bson import RawBSONDocument
 from bson.son import SON
+from pymongo.client_session import _validate_session_write_concern
 from pymongo.common import (validate_is_mapping,
                             validate_is_document_type,
                             validate_ok_for_replace,
                             validate_ok_for_update)
+from pymongo.helpers import _RETRYABLE_ERROR_CODES
 from pymongo.collation import validate_collation_or_none
 from pymongo.errors import (BulkWriteError,
                             ConfigurationError,
-                            ConnectionFailure,
                             InvalidOperation,
-                            OperationFailure,
-                            ServerSelectionTimeoutError)
+                            OperationFailure)
 from pymongo.message import (_INSERT, _UPDATE, _DELETE,
                              _do_batched_insert,
-                             _do_batched_write_command,
                              _randint,
-                             _BulkWriteContext)
+                             _BulkWriteContext,
+                             _EncryptedBulkWriteContext)
+from pymongo.read_preferences import ReadPreference
 from pymongo.write_concern import WriteConcern
 
 
@@ -88,46 +91,53 @@ class _Run(object):
         self.ops.append(operation)
 
 
-def _merge_command(run, full_result, results):
-    """Merge a group of results from write commands into the full result.
+def _merge_command(run, full_result, offset, result):
+    """Merge a write command result into the full bulk result.
     """
-    for offset, result in results:
+    affected = result.get("n", 0)
 
-        affected = result.get("n", 0)
+    if run.op_type == _INSERT:
+        full_result["nInserted"] += affected
 
-        if run.op_type == _INSERT:
-            full_result["nInserted"] += affected
+    elif run.op_type == _DELETE:
+        full_result["nRemoved"] += affected
 
-        elif run.op_type == _DELETE:
-            full_result["nRemoved"] += affected
+    elif run.op_type == _UPDATE:
+        upserted = result.get("upserted")
+        if upserted:
+            n_upserted = len(upserted)
+            for doc in upserted:
+                doc["index"] = run.index(doc["index"] + offset)
+            full_result["upserted"].extend(upserted)
+            full_result["nUpserted"] += n_upserted
+            full_result["nMatched"] += (affected - n_upserted)
+        else:
+            full_result["nMatched"] += affected
+        full_result["nModified"] += result["nModified"]
 
-        elif run.op_type == _UPDATE:
-            upserted = result.get("upserted")
-            if upserted:
-                n_upserted = len(upserted)
-                for doc in upserted:
-                    doc["index"] = run.index(doc["index"] + offset)
-                full_result["upserted"].extend(upserted)
-                full_result["nUpserted"] += n_upserted
-                full_result["nMatched"] += (affected - n_upserted)
-            else:
-                full_result["nMatched"] += affected
-            full_result["nModified"] += result["nModified"]
+    write_errors = result.get("writeErrors")
+    if write_errors:
+        for doc in write_errors:
+            # Leave the server response intact for APM.
+            replacement = doc.copy()
+            idx = doc["index"] + offset
+            replacement["index"] = run.index(idx)
+            # Add the failed operation to the error document.
+            replacement[_UOP] = run.ops[idx]
+            full_result["writeErrors"].append(replacement)
 
-        write_errors = result.get("writeErrors")
-        if write_errors:
-            for doc in write_errors:
-                # Leave the server response intact for APM.
-                replacement = doc.copy()
-                idx = doc["index"] + offset
-                replacement["index"] = run.index(idx)
-                # Add the failed operation to the error document.
-                replacement[_UOP] = run.ops[idx]
-                full_result["writeErrors"].append(replacement)
+    wc_error = result.get("writeConcernError")
+    if wc_error:
+        full_result["writeConcernErrors"].append(wc_error)
 
-        wc_error = result.get("writeConcernError")
-        if wc_error:
-            full_result["writeConcernErrors"].append(wc_error)
+
+def _raise_bulk_write_error(full_result):
+    """Raise a BulkWriteError from the full bulk api result.
+    """
+    if full_result["writeErrors"]:
+        full_result["writeErrors"].sort(
+            key=lambda error: error["index"])
+    raise BulkWriteError(full_result)
 
 
 class _Bulk(object):
@@ -142,17 +152,25 @@ class _Bulk(object):
                 document_class=dict))
         self.ordered = ordered
         self.ops = []
-        self.name = "%s.%s" % (collection.database.name, collection.name)
-        self.namespace = collection.database.name + '.$cmd'
         self.executed = False
         self.bypass_doc_val = bypass_document_validation
         self.uses_collation = False
         self.uses_array_filters = False
+        self.uses_hint = False
         self.is_retryable = True
         self.retrying = False
+        self.started_retryable_write = False
         # Extra state so that we know where to pick up on a retry attempt.
         self.current_run = None
 
+    @property
+    def bulk_ctx_class(self):
+        encrypter = self.collection.database.client._encrypter
+        if encrypter and not encrypter._bypass_auto_encryption:
+            return _EncryptedBulkWriteContext
+        else:
+            return _BulkWriteContext
+
     def add_insert(self, document):
         """Add an insert document to the list of ops.
         """
@@ -163,7 +181,7 @@ class _Bulk(object):
         self.ops.append((_INSERT, document))
 
     def add_update(self, selector, update, multi=False, upsert=False,
-                   collation=None, array_filters=None):
+                   collation=None, array_filters=None, hint=None):
         """Create an update document and add it to the list of ops.
         """
         validate_ok_for_update(update)
@@ -176,13 +194,16 @@ class _Bulk(object):
         if array_filters is not None:
             self.uses_array_filters = True
             cmd['arrayFilters'] = array_filters
+        if hint is not None:
+            self.uses_hint = True
+            cmd['hint'] = hint
         if multi:
             # A bulk_write containing an update_many is not retryable.
             self.is_retryable = False
         self.ops.append((_UPDATE, cmd))
 
     def add_replace(self, selector, replacement, upsert=False,
-                    collation=None):
+                    collation=None, hint=None):
         """Create a replace document and add it to the list of ops.
         """
         validate_ok_for_replace(replacement)
@@ -192,9 +213,12 @@ class _Bulk(object):
         if collation is not None:
             self.uses_collation = True
             cmd['collation'] = collation
+        if hint is not None:
+            self.uses_hint = True
+            cmd['hint'] = hint
         self.ops.append((_UPDATE, cmd))
 
-    def add_delete(self, selector, limit, collation=None):
+    def add_delete(self, selector, limit, collation=None, hint=None):
         """Create a delete document and add it to the list of ops.
         """
         cmd = SON([('q', selector), ('limit', limit)])
@@ -202,6 +226,9 @@ class _Bulk(object):
         if collation is not None:
             self.uses_collation = True
             cmd['collation'] = collation
+        if hint is not None:
+            self.uses_hint = True
+            cmd['hint'] = hint
         if limit == _DELETE_ALL:
             # A bulk_write containing a delete_many is not retryable.
             self.is_retryable = False
@@ -235,9 +262,13 @@ class _Bulk(object):
 
     def _execute_command(self, generator, write_concern, session,
                          sock_info, op_id, retryable, full_result):
-        if sock_info.max_wire_version < 5 and self.uses_collation:
-            raise ConfigurationError(
-                'Must be connected to MongoDB 3.4+ to use a collation.')
+        if sock_info.max_wire_version < 5:
+            if self.uses_collation:
+                raise ConfigurationError(
+                    'Must be connected to MongoDB 3.4+ to use a collation.')
+            if self.uses_hint:
+                raise ConfigurationError(
+                    'Must be connected to MongoDB 3.4+ to use hint.')
         if sock_info.max_wire_version < 6 and self.uses_array_filters:
             raise ConfigurationError(
                 'Must be connected to MongoDB 3.6+ to use arrayFilters.')
@@ -256,39 +287,45 @@ class _Bulk(object):
         while run:
             cmd = SON([(_COMMANDS[run.op_type], self.collection.name),
                        ('ordered', self.ordered)])
-            if write_concern.document:
+            if not write_concern.is_server_default:
                 cmd['writeConcern'] = write_concern.document
             if self.bypass_doc_val and sock_info.max_wire_version >= 4:
                 cmd['bypassDocumentValidation'] = True
-            if session:
-                cmd['lsid'] = session._use_lsid()
-            bwc = _BulkWriteContext(db_name, cmd, sock_info, op_id,
-                                    listeners, session)
+            bwc = self.bulk_ctx_class(
+                db_name, cmd, sock_info, op_id, listeners, session,
+                run.op_type, self.collection.codec_options)
 
-            results = []
             while run.idx_offset < len(run.ops):
-                if session and retryable:
-                    cmd['txnNumber'] = session._transaction_id()
+                if session:
+                    # Start a new retryable write unless one was already
+                    # started for this command.
+                    if retryable and not self.started_retryable_write:
+                        session._start_retryable_write()
+                        self.started_retryable_write = True
+                    session._apply_to(cmd, retryable, ReadPreference.PRIMARY)
                 sock_info.send_cluster_time(cmd, session, client)
-                check_keys = run.op_type == _INSERT
                 ops = islice(run.ops, run.idx_offset, None)
-                # Run as many ops as possible.
-                request_id, msg, to_send = _do_batched_write_command(
-                    self.namespace, run.op_type, cmd, ops, check_keys,
-                    self.collection.codec_options, bwc)
-                if not to_send:
-                    raise InvalidOperation("cannot do an empty bulk write")
-                result = bwc.write_command(request_id, msg, to_send)
-                client._receive_cluster_time(result, session)
-                results.append((run.idx_offset, result))
+                # Run as many ops as possible in one command.
+                result, to_send = bwc.execute(ops, client)
+
+                # Retryable writeConcernErrors halt the execution of this run.
+                wce = result.get('writeConcernError', {})
+                if wce.get('code', 0) in _RETRYABLE_ERROR_CODES:
+                    # Synthesize the full bulk result without modifying the
+                    # current one because this write operation may be retried.
+                    full = copy.deepcopy(full_result)
+                    _merge_command(run, full, run.idx_offset, result)
+                    _raise_bulk_write_error(full)
+
+                _merge_command(run, full_result, run.idx_offset, result)
                 # We're no longer in a retry once a command succeeds.
                 self.retrying = False
+                self.started_retryable_write = False
+
                 if self.ordered and "writeErrors" in result:
                     break
                 run.idx_offset += len(to_send)
 
-            _merge_command(run, full_result, results)
-
             # We're supposed to continue if errors are
             # at the write concern level (e.g. wtimeout)
             if self.ordered and full_result['writeErrors']:
@@ -323,10 +360,7 @@ class _Bulk(object):
                 self.is_retryable, retryable_bulk, s, self)
 
         if full_result["writeErrors"] or full_result["writeConcernErrors"]:
-            if full_result['writeErrors']:
-                full_result['writeErrors'].sort(
-                    key=lambda error: error['index'])
-            raise BulkWriteError(full_result)
+            _raise_bulk_write_error(full_result)
         return full_result
 
     def execute_insert_no_results(self, sock_info, run, op_id, acknowledged):
@@ -341,12 +375,64 @@ class _Bulk(object):
         db = self.collection.database
         bwc = _BulkWriteContext(
             db.name, command, sock_info, op_id, db.client._event_listeners,
-            session=None)
+            None, _INSERT, self.collection.codec_options)
         # Legacy batched OP_INSERT.
         _do_batched_insert(
             self.collection.full_name, run.ops, True, acknowledged, concern,
             not self.ordered, self.collection.codec_options, bwc)
 
+    def execute_op_msg_no_results(self, sock_info, generator):
+        """Execute write commands with OP_MSG and w=0 writeConcern, unordered.
+        """
+        db_name = self.collection.database.name
+        client = self.collection.database.client
+        listeners = client._event_listeners
+        op_id = _randint()
+
+        if not self.current_run:
+            self.current_run = next(generator)
+        run = self.current_run
+
+        while run:
+            cmd = SON([(_COMMANDS[run.op_type], self.collection.name),
+                       ('ordered', False),
+                       ('writeConcern', {'w': 0})])
+            bwc = self.bulk_ctx_class(
+                db_name, cmd, sock_info, op_id, listeners, None,
+                run.op_type, self.collection.codec_options)
+
+            while run.idx_offset < len(run.ops):
+                ops = islice(run.ops, run.idx_offset, None)
+                # Run as many ops as possible.
+                to_send = bwc.execute_unack(ops, client)
+                run.idx_offset += len(to_send)
+            self.current_run = run = next(generator, None)
+
+    def execute_command_no_results(self, sock_info, generator):
+        """Execute write commands with OP_MSG and w=0 WriteConcern, ordered.
+        """
+        full_result = {
+            "writeErrors": [],
+            "writeConcernErrors": [],
+            "nInserted": 0,
+            "nUpserted": 0,
+            "nMatched": 0,
+            "nModified": 0,
+            "nRemoved": 0,
+            "upserted": [],
+        }
+        # Ordered bulk writes have to be acknowledged so that we stop
+        # processing at the first error, even when the application
+        # specified unacknowledged writeConcern.
+        write_concern = WriteConcern()
+        op_id = _randint()
+        try:
+            self._execute_command(
+                generator, write_concern, None,
+                sock_info, op_id, False, full_result)
+        except OperationFailure:
+            pass
+
     def execute_no_results(self, sock_info, generator):
         """Execute all operations, returning no results (w=0).
         """
@@ -356,10 +442,20 @@ class _Bulk(object):
         if self.uses_array_filters:
             raise ConfigurationError(
                 'arrayFilters is unsupported for unacknowledged writes.')
-        # Cannot have both unacknowledged write and bypass document validation.
+        if self.uses_hint:
+            raise ConfigurationError(
+                'hint is unsupported for unacknowledged writes.')
+        # Cannot have both unacknowledged writes and bypass document validation.
         if self.bypass_doc_val and sock_info.max_wire_version >= 4:
             raise OperationFailure("Cannot set bypass_document_validation with"
                                    " unacknowledged write concern")
+
+        # OP_MSG
+        if sock_info.max_wire_version > 5:
+            if self.ordered:
+                return self.execute_command_no_results(sock_info, generator)
+            return self.execute_op_msg_no_results(sock_info, generator)
+
         coll = self.collection
         # If ordered is True we have to send GLE or use write
         # commands so we can abort on the first error.
@@ -416,8 +512,8 @@ class _Bulk(object):
             raise InvalidOperation('Bulk operations can '
                                    'only be executed once.')
         self.executed = True
-        write_concern = (WriteConcern(**write_concern) if
-                         write_concern else self.collection.write_concern)
+        write_concern = write_concern or self.collection.write_concern
+        session = _validate_session_write_concern(session, write_concern)
 
         if self.ordered:
             generator = self.gen_ordered()
@@ -426,7 +522,7 @@ class _Bulk(object):
 
         client = self.collection.database.client
         if not write_concern.acknowledged:
-            with client._socket_for_writes() as sock_info:
+            with client._socket_for_writes(session) as sock_info:
                 self.execute_no_results(sock_info, generator)
         else:
             return self.execute_command(generator, write_concern, session)
@@ -608,5 +704,5 @@ class BulkOperationBuilder(object):
             execution.
         """
         if write_concern is not None:
-            validate_is_mapping("write_concern", write_concern)
+            write_concern = WriteConcern(**write_concern)
         return self.__bulk.execute(write_concern, session=None)
